require File.join(File.dirname(__FILE__), 'certapi')
require 'dalli'

module CertificateHelpers
  # get Authorized User info.  This implementation may change
  # so that the authorized user information is more easily accessible
  # among a broad range of applications, but for now this is what works.
  def user(env)
    ConsoleAuth.get_user(env)
  end

  def csr_exists?(certname)
    current_cert_status = CertAPI.cert_info certname
    (! [*current_cert_status].empty?) and current_cert_status['state'] == 'requested'
  end

  def get_csrs
    begin
      CertAPI.csrs
    rescue CertAPI::ApiUnauthorizedException
      halt 403, "Permissions error"
    rescue
      halt 500, "API failure"
    end
  end

  def real_and_cached_csrs
    cached = get_all_cached_nodes
    get_csrs.concat(cached.values).sort_by { |cert| cert['name'] }
  end

  def single_sign(certname, logger)
    csr = CertAPI.cert_info(certname)
    api_response = logger.log_action(user(env), certname, "Add") do
      CertAPI.add(certname)
    end
    unless is_successful?(api_response.code)
      halt 403, "A pending CSR for #{certname} was found, but the CA failed to sign it." +
                " The CA responded with the message '#{api_response.body}'"
    end
    csr['state'] = 'signed'
    add_to_cache(certname, csr)
    api_response
  end

  def multi_sign(certnames, logger)
    result = {}
    certnames.each do |certname|
      unless csr_exists?(certname)
        result[certname] = {
          :result  => 'fail',
          :message => "#{certname} could not be accepted because its certificate " +
                      "request has been moved or modified."
        }
      else
        csr = CertAPI.cert_info(certname)
        api_response = logger.log_action(user(env), certname, "Add") do
          CertAPI.add(certname)
        end
        if is_successful?(api_response.code)
          csr['state'] = 'signed'
          add_to_cache(certname, csr)
          result[certname] = {
            :result => 'success'
          }
        else
          result[certname] = {
            :result  => 'fail',
            :message => "A pending CSR for #{certname} was found, but the CA failed to sign it." +
                        " The CA responded with the message '#{api_response.body}'"
          }
        end
      end
    end
    result
  end

  def single_reject(certname, logger)
    # The caching case for the reject/remove is slightly more complicated
    # than the other cases because once the csr has been rejected, you
    # can't request its state after rejected. To get around that we have
    # to make a temporary version of the csr, set its state to rejected
    # and then, if the call to reject succeeds, store that temporary
    # version in the cache
    cert = CertAPI.cert_info(certname)
    cert['state'] = 'rejected'
    api_response = logger.log_action(user(env), certname, "Remove") do
      CertAPI.remove(certname)
    end
    unless is_successful?(api_response.code)
      halt 403, "A pending CSR for #{certname} was found, but the CA failed to sign it." +
                " The CA responded with the message '#{api_response.body}'"
    end
    add_to_cache(certname, cert)
    api_response
  end

  def multi_reject(certnames, logger)
    result = {}
    certnames.each do |certname|
      unless csr_exists?(certname)
        result[certname] = {
          :result  => 'fail',
          :message => "#{certname} could not be rejected because its certificate " +
                      "request has been moved or modified."
        }
      else
        cert = CertAPI.cert_info(certname)
        cert['state'] = 'rejected'
        api_response = logger.log_action(user(env), certname, "Remove") do
          CertAPI.remove(certname)
        end
        if is_successful?(api_response.code)
          add_to_cache(certname, cert)
          result[certname] = {
            :result => 'success'
          }
        else
          result[certname] = {
            :result  => 'fail',
            :message => "A pending CSR for #{certname} was found, but the CA failed to reject it." +
                        " The CA responded with the message '#{api_response.body}'"
          }
        end
      end
    end
    result
  end

  # this is intentionally not named successful? so as to not overwrite the
  # existing Rack method successful?
  def is_successful?(status)
    status.to_i < 300 && status.to_i >= 200
  end

  def memcache_client
    @memcache_client ||= Dalli::Client.new('127.0.0.1:11211',
                                           :compression => true,
                                           :namespace => 'certmgr',
                                           :expires_in => cache_expiration)
  end

  def get_all_cached_nodes
    purge_memcache_keys || {}
  end

  def add_to_cache(node, data)
    memcache_client.multi do
      add_node_to_array node
      memcache_client.set node, data
    end
  end

  def purge_from_cache(node)
    memcache_client.delete node
  end

  def multi_purge(certnames, logger)
    result = {}
    certnames.each do |certname|
      # logging because this action will affect the display for other users
      response_code = logger.log_action(user(env), certname, "Cache purge") do
        purge_from_cache(certname) ? 200 : 500
      end

      if is_successful?(response_code)
        result[certname] = {:result => 'success'}
      else
        result[certname] = {:result => 'fail'}
      end
    end
    result
  end

  def purge_memcache_keys
    memcache_client.multi do
      nodes = memcache_client.get node_array_key
      if nodes
        all_node_data = memcache_client.get_multi nodes
        nodes = nodes.delete_if { |n| all_node_data[n].nil? }
        memcache_client.set node_array_key, nodes
        all_node_data.delete_if { |key, value| value == nil }
      end
    end
  end

  def add_node_to_array(node)
    nodes = memcache_client.get node_array_key
    nodes ||= []
    nodes << node
    memcache_client.set node_array_key, nodes
  end

  def node_array_key
    :actioned_node_list
  end

  def cache_expiration
    # 1 day in seconds
    60 * 60 * 24
  end
end
